(**************************************************************************)
(*                                                                        *)
(*  Input module for ToUTF8 utility                                       *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE InStream;

        (********************************************************)
        (*                                                      *)
        (*            Input stream for ToUTF8 utility           *)
        (*                                                      *)
        (*   The main purpose of this module is to reduce a     *)
        (*   variety of Unicode encodings, as defined by the    *)
        (*   byte order mark, to a uniform UTF-8 stream.  Files *)
        (*   using a non-Unicode character set have their       *)
        (*   contents passed through unchanged.                 *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            13 July 2024                    *)
        (*  Last edited:        15 July 2024                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, ADR;

FROM Arith64 IMPORT
    (* proc *)  Add64LE;

FROM FileOps IMPORT
    (* type *)  ChanId, FilePos,
    (* proc *)  StartPosition, CurrentPosition,
                SetPosition, ReadRaw;

FROM Heap IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM LowLevel IMPORT
    (* proc *)  Copy;

(************************************************************************)

CONST
    BufferSize = 4096;
    Nul = CHR(0);

TYPE
    EncodingType = (byte, twobyte, fourbyte);

    (*    cid            channel ID of the opened file                  *)
    (*    offset         offset within CurrentBuffer to the next        *)
    (*                    character to be read in a sequential read.    *)
    (*    encoding       flag to label a UTF-16 or UTF-32 file.         *)
    (*    BigEndian      Meaningful only if encoding <> byte.           *)

    Stream = POINTER TO StateRecord;
    StateRecord = RECORD
                      cid: ChanId;
                      LastPos: FilePos;
                      encoding: EncodingType;
                      BigEndian: BOOLEAN;
                  END (*RECORD*);

(************************************************************************)
(*                      OUTPUT OF A UTF8 STRING                         *)
(************************************************************************)

PROCEDURE StoreUTF8 (code: CARDINAL;  VAR (*INOUT*) buffer: ARRAY OF CHAR;
                            VAR (*INOUT*) pos: CARDINAL);

    (* Stores the UTF8 encoding of code at buffer[pos], updates pos. *)

    BEGIN
        IF code = 0 THEN
            buffer[pos] := '?';  INC(pos);
        ELSIF code < 080H THEN
            buffer[pos] := CHR(code);  INC(pos);
        ELSE
            IF code < 0800H THEN
                buffer[pos] := CHR(0C0H + code DIV 64);  INC(pos);
            ELSE
                IF code < 01000H THEN
                    buffer[pos] := CHR(0E0H + code DIV 4096);  INC(pos);
                ELSE
                    buffer[pos] := CHR(0F0H + code DIV 262144);  INC(pos);
                    code := code MOD 262144;
                    buffer[pos] := CHR(080H + code DIV 4096);  INC(pos);
                END (*IF*);
                code := code MOD 4096;
                buffer[pos] := CHR(080H + code DIV 64);  INC(pos);
            END (*IF*);
            buffer[pos] := CHR(080H + code MOD 64);  INC(pos);
        END (*IF*);
    END StoreUTF8;

(************************************************************************)
(*               DEALING WITH 16-bit and 32-bit UNICODE                 *)
(************************************************************************)

PROCEDURE ConvertToUTF8 (VAR (*INOUT*) data: ARRAY OF CHAR;
                             encoding: EncodingType;
                             BigEndian: BOOLEAN;
                             VAR (*INOUT*) bytecount: CARDINAL);

    (* Converts bytecount bytes from UTF-32 or UTF-16 to UTF-8, in      *)
    (* place, and updates bytecount to reflect the result.  If          *)
    (* bytecount is not a multiple of the basic encoding unit on entry  *)
    (* (which shouldn't happen), some trailing bytes of input are       *)
    (* dropped.                                                         *)

    VAR incount: CARDINAL;

    (********************************************************************)

    PROCEDURE GetTwoBytes(): CARDINAL;

        (* Fetches a 16-bit value from the data array. *)

        VAR result: CARDINAL;

        BEGIN
            result := ORD(data[incount]);  INC (incount);
            IF BigEndian THEN
                result := 256*result + ORD(data[incount]);
            ELSE
                result := result + 256*ORD(data[incount]);
            END (*IF*);
            INC (incount);
            RETURN result;
        END GetTwoBytes;

    (********************************************************************)

    PROCEDURE GetFourBytes(): CARDINAL;

        (* Fetches a 32-bit value from the data array. *)

        VAR result: CARDINAL;

        BEGIN
            result := GetTwoBytes();
            IF BigEndian THEN
                result := 65536*result + GetTwoBytes();
            ELSE
                result := result + 65536*GetTwoBytes();
            END (*IF*);
            RETURN result;
        END GetFourBytes;

    (********************************************************************)

    VAR outcount, code: CARDINAL;
        result: ARRAY [0..3*BufferSize/2-1] OF CHAR;

        (* The size of the result array is based on the worst case,     *)
        (* where a 16-bit UTF16 code translates to a 3-byte result.     *)
        (* In that (rare) worst case we might end up truncating the     *)
        (* result, but that's OK since the procedure that calls us      *)
        (* will pick up what's missing on the next call, if needed.     *)

    BEGIN
        IF encoding = byte THEN
            RETURN;       (* No transformation needed *)
        END (*IF*);

        IF ODD(bytecount) THEN
            DEC (bytecount);
        END (*IF*);
        incount := 0;  outcount := 0;
        WHILE incount < bytecount DO
            IF encoding = fourbyte THEN
                code := GetFourBytes();
            ELSE
                code := GetTwoBytes();
                IF (code >= 0D800H) AND (code <= 0DFFFFH) THEN
                    (* Get the second word of a two-word UTF16 code. *)
                    code := 1024*(code MOD 1024) + (GetTwoBytes() MOD 1024);
                END (*IF*);
            END (*IF*);

            (* Now that we have the Unicode value, translate it to UTF8. *)

            StoreUTF8 (code, result, outcount);

        END (*WHILE*);

        IF outcount > 0 THEN
            Copy (ADR(result), ADR(data), outcount);
        END (*IF*);
        bytecount := outcount;

    END ConvertToUTF8;

(************************************************************************)
(*                           OPEN/CLOSE/ETC                             *)
(************************************************************************)

PROCEDURE OpenStream (cid: ChanId): Stream;

    (* Collects some information about the input, in particular *)
    (* checking the Byte Order Mark if present.                 *)

    VAR result: Stream;
        data: ARRAY [0..3] OF CHAR;
        NumberRead, BOMoffset: CARDINAL;

    BEGIN
        NEW (result);
        result^.cid := cid;
        result^.LastPos := StartPosition(cid);
        result^.encoding := byte;
        result^.BigEndian := FALSE;
        BOMoffset := 0;

        (* Still have to interpret the BOM.  It can be up to four bytes long. *)

        ReadRaw (cid, data, 4, NumberRead);

        (* Check for Unicode.  We work out which variant we have *)
        (* by looking for either a byte order mark or a pattern  *)
        (* of Nul bytes.                                         *)

        IF data[0] = CHR(0) THEN
            result^.BigEndian := TRUE;
            IF data[1] = CHR(0) THEN
                result^.encoding := fourbyte;
                IF (data[2] = CHR(0FEH)) AND (data[3] = CHR(0FFH)) THEN
                    BOMoffset := 4;
                END (*IF*);
            ELSE
                result^.encoding := twobyte;
            END (*IF*);
        ELSIF data[1] = CHR(0) THEN
            result^.encoding := twobyte;  result^.BigEndian := FALSE;
        ELSIF (data[0] = CHR(0EFH)) AND (data[1] = CHR(0BBH))
                                            AND (data[2] = CHR(0BFH)) THEN
            result^.encoding := byte;
            BOMoffset := 3;
        ELSIF (data[0] = CHR(0FFH)) AND (data[1] = CHR(0FEH)) THEN
            result^.encoding := twobyte;     (* tentatively, see below *)
            result^.BigEndian := FALSE;
            BOMoffset := 2;
        ELSIF (data[0] = CHR(0FEH)) AND (data[1] = CHR(0FFH)) THEN
            result^.encoding := twobyte;  result^.BigEndian := TRUE;
            BOMoffset := 2;
        END (*IF*);

        (* We still have to check the case where third and fourth bytes are zero. *)

        IF (data[2] = CHR(0)) AND (data[3] = CHR(0)) THEN
            result^.encoding := fourbyte;
            IF BOMoffset > 0 THEN
                BOMoffset := 4;
            END (*IF*);
        END (*IF*);

        (* Position the read position just after the BOM. *)

        SetPosition (cid, StartPosition(cid));
        IF BOMoffset > 0 THEN
            ReadRaw (cid, data, BOMoffset, NumberRead);
        END (*IF*);

        RETURN result;
    END OpenStream;

(************************************************************************)

PROCEDURE CloseStream (S: Stream);

    (* Drops the use of this stream.  *)

    BEGIN
        DISPOSE (S);
    END CloseStream;

(************************************************************************)
(*                              INPUT                                   *)
(************************************************************************)

PROCEDURE Adjustsize (S: Stream;  VAR (*INOUT*) size: CARDINAL);

    (* Size is the intended number of bytes to be read.  Reduces size,  *)
    (* if necessary, to be compatible with S^.encoding.                 *)

    BEGIN
        IF S^.encoding = twobyte THEN
            (* Reduce size to allow for expansion when UTF16 codes  *)
            (* are translated to UTF8.  In all other cases there    *)
            (* will be no contraction.                              *)
            size := 2*size DIV 3;
            IF ODD(size) THEN DEC(size) END(*IF*);
        ELSIF S^.encoding = fourbyte THEN
            DEC (size, size MOD 4);
        END (*IF*);
    END Adjustsize;

(************************************************************************)

PROCEDURE SReadRaw (S: Stream;  VAR (*OUT*) data: ARRAY OF CHAR;
                     limit: CARDINAL;  VAR (*OUT*) NumberRead: CARDINAL);

    (* Reads a buffer-full of information from a file. *)

    BEGIN
        S^.LastPos := CurrentPosition(S^.cid);
        Adjustsize (S, limit);
        ReadRaw (S^.cid, data, limit, NumberRead);
        data[NumberRead] := Nul;
        IF S^.encoding <> byte THEN
            ConvertToUTF8 (data, S^.encoding, S^.BigEndian, NumberRead);
        END (*IF*);
    END SReadRaw;

(************************************************************************)

PROCEDURE OldSReadLine (S: Stream;  VAR (*OUT*) line: ARRAY OF CHAR);

    (* Reads a line of text from a file.  *)

    (********************************************************************)

    PROCEDURE GetVal (pos, N: CARDINAL;  BigEndian: BOOLEAN): CARDINAL;

        (* Picks up a value at line[pos] that is N bytes long.   *)
        (* Note that the caller's value of pos is not altered.   *)

        VAR val: RECORD
                     CASE :BOOLEAN OF
                            FALSE: byte: ARRAY [0..3] OF CARD8;
                          |
                            TRUE:  word: CARDINAL;
                     END (*CASE*);
                 END (*RECORD*);
            j: CARDINAL;

        BEGIN
            val.word := 0;
            IF BigEndian THEN
                FOR j := N-1 TO 0 BY -1 DO
                    val.byte[j] := ORD(line[pos]);  INC(pos);
                END (*FOR*);
            ELSE
                FOR j := 0 TO N-1 DO
                    val.byte[j] := ORD(line[pos]);  INC(pos);
                END (*FOR*);
            END (*IF*);
            RETURN val.word;
        END GetVal;

    (********************************************************************)

    CONST CR = 13;  LF = 10;

    VAR pos, size, step, k, NumberRead: CARDINAL;  found: BOOLEAN;
        filepos: FilePos;

    BEGIN
        S^.LastPos := CurrentPosition(S^.cid);
        size := HIGH(line);
        Adjustsize (S, size);

        IF S^.encoding = fourbyte THEN
            step := 4;
        ELSIF S^.encoding = twobyte THEN
            step := 2;
        ELSE
            step := 1;
        END (*IF*);

        ReadRaw (S^.cid, line, HIGH(line), NumberRead);

        (* Search for first line feed.  We can't use Strings.FindNext   *)
        (* to do this because of the possible presence of zero bytes.   *)

        pos := 0;  found := FALSE;
        WHILE (NOT found) AND (pos <= NumberRead-step) DO
            found := GetVal (pos, step, S^.BigEndian) = LF;
            IF NOT found THEN INC(pos, step) END(*IF*);
        END (*WHILE*);

        (* Step back to preceding CR, if present. *)

        IF found THEN
            IF pos >= step THEN
                IF GetVal (pos-step, step, S^.BigEndian) = CR THEN
                    DEC (pos, step);
                END (*IF*);
            END (*IF*);
        END (*IF*);
        line[pos] := Nul;

        (* Now we have to set the file pointer to the character *)
        (* corresponding to line[pos].                          *)

        filepos := S^.LastPos;
        Add64LE (filepos, pos);
        SetPosition (S^.cid, filepos);
        IF S^.encoding <> byte THEN
            ConvertToUTF8 (line, S^.encoding, S^.BigEndian, k);
        END (*IF*);

    END OldSReadLine;

(************************************************************************)

PROCEDURE SReadLine (S: Stream;  VAR (*OUT*) line: ARRAY OF CHAR);

    (* Reads a line of text from a file.  *)

    (********************************************************************)

    PROCEDURE GetVal (code, N: CARDINAL;  BigEndian: BOOLEAN): CARDINAL;

        (* Translates code, which is N bytes long, to CARDINAL.  We     *)
        (* assume that the caller has set unused bytes to zero.         *)

        VAR inval, outval: RECORD
                               CASE :BOOLEAN OF
                                      FALSE: byte: ARRAY [0..3] OF CARD8;
                                    |
                                      TRUE:  word: CARDINAL;
                               END (*CASE*);
                           END (*RECORD*);
            j: CARDINAL;

        BEGIN
            IF N = 1 THEN
                outval.word := code;
            ELSE
                inval.word := code;
                IF BigEndian THEN
                    FOR j := 0 TO N-1 DO
                        outval.byte[j] := inval.byte[N-j-1];
                    END (*FOR*);
                END (*IF*);
            END (*IF*);
            RETURN outval.word;
        END GetVal;

    (********************************************************************)

    CONST CR = 13;  LF = 10;

    VAR pos, charsize, NumberRead, widech, val: CARDINAL;

    BEGIN
        S^.LastPos := CurrentPosition(S^.cid);

        IF S^.encoding = fourbyte THEN
            charsize := 4;
        ELSIF S^.encoding = twobyte THEN
            charsize := 2;
        ELSE
            charsize := 1;
        END (*IF*);

        (* We exit the following loop at EOF or after finding LF. *)

        pos := 0;
        LOOP
            widech := 0;
            ReadRaw (S^.cid, widech, charsize, NumberRead);
            IF NumberRead = 0 THEN EXIT(*LOOP*) END(*IF*);
            val := GetVal (widech, charsize, S^.BigEndian);
            IF val = LF THEN EXIT(*LOOP*);
            ELSIF val <> CR THEN
                StoreUTF8 (val, line, pos);
            END (*IF*);
        END (*LOOP*);
        line[pos] := Nul;

    END SReadLine;

(************************************************************************)

PROCEDURE StepBack (S: Stream);

    (* Moves the file pointer back to where it was before   *)
    (* the last SReadLine or SReadRaw.                      *)

    BEGIN
        SetPosition (S^.cid, S^.LastPos);
    END StepBack;

(************************************************************************)

END InStream.

