MODULE ToUTF8;

        (********************************************************)
        (*                                                      *)
        (*   Converts a GEDCOM file to the UTF8 character set.  *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            30 December 2024                *)
        (*  Last edited:        30 December 2024                *)
        (*  Status:             Complete, not fully tested      *)
        (*                                                      *)
        (*  Don't yet handle wildcards in input parameter       *)
        (*                                                      *)
        (*    This program takes a single parameter, which can  *)
        (*    be the name of a GEDCOM file, or a wildcard       *)
        (*    specifying one or more files.                     *)
        (*                                                      *)
        (********************************************************)


IMPORT IOChan, TextIO, FileSys, Strings;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM Transl8 IMPORT
    (* type *)  InState,
    (* proc *)  StartReading, TranslateStream, StopReading;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilenameString, DirectoryEntry,
    (* proc *)  OpenNewFile1, OpenOldFile, CloseFile, DeleteFile, MoveFile,
    (* proc *)  SetPosition, StartPosition, ReadRaw, WriteRaw,
                ReadLine, FWriteString, FWriteLn,
                FirstDirEntry, NextDirEntry, DirSearchDone;

(********************************************************************************)

CONST
    Debugging = FALSE;
    DefaultInput = "testANSEL";
    Nul = CHR(0);  CtrlZ = CHR(26);

VAR
    (* Variable used in constructing a unique file name. *)

    NextName: ARRAY [0..7] OF CHAR;

(********************************************************************************)
(*                          GET THE DATABASE NAME                               *)
(********************************************************************************)

PROCEDURE GetArgument (VAR (*OUT*) DatabaseName: FilenameString);

    (* Picks up the argument - a single string - to the program. *)

    VAR cid: IOChan.ChanId;  j: CARDINAL;

    BEGIN
        IF Debugging THEN
            DatabaseName := DefaultInput;
        ELSE
            cid := ArgChan();
            IF IsArgPresent() THEN
                TextIO.ReadString (cid, DatabaseName);
                j := LENGTH (DatabaseName);
                REPEAT
                    DEC (j);
                UNTIL (j = 0) OR (DatabaseName[j] <> ' ');
                DatabaseName[j+1] := Nul;
            ELSE
                DatabaseName := "";
            END (*IF*);
        END (*IF*);

    END GetArgument;

(************************************************************************)
(*                    CREATING A UNIQUE FILENAME                        *)
(************************************************************************)

PROCEDURE MakeUniqueName (VAR (*OUT*) name: ARRAY OF CHAR);

    (* Generates a unique 8-character string.  The argument must of     *)
    (* course be big enough to take at least 8 characters.              *)

    (********************************************************************)

    PROCEDURE Increment (N: CARDINAL);

        (* Increments NextName[N], with carry as appropriate. *)

        BEGIN
            IF NextName[N] = '9' THEN
                NextName[N] := 'A';
            ELSIF NextName[N] = 'Z' THEN
                NextName[N] := '0';
                IF N > 0 THEN
                    Increment (N-1);
                END (*IF*);
            ELSE
                INC (NextName[N]);
            END (*IF*);
        END Increment;

    (********************************************************************)

    BEGIN
        Strings.Assign (NextName, name);
        Increment (7);
    END MakeUniqueName;

(************************************************************************)

PROCEDURE MakeNewFilename (VAR (*OUT*) NewName: ARRAY OF CHAR);

    (* Creates a file name of the form BaseNamexxxtail, where xxx is    *)
    (* chosen such that a file of that name does not already exist.     *)
    (* Note that BaseName and tail can include punctuation.             *)

    BEGIN
        REPEAT
            MakeUniqueName (NewName);
        UNTIL NOT FileSys.Exists(NewName);
    END MakeNewFilename;

(************************************************************************)

PROCEDURE OpenNewOutputFile (VAR (*OUT*) NewName: FilenameString): ChanId;

    (* Creates a file name that does not already exist, and *)
    (* opens that file.                                     *)

    VAR cid: ChanId;  duplication: BOOLEAN;

    BEGIN
        REPEAT
            MakeNewFilename (NewName);
            cid := OpenNewFile1 (NewName, duplication);
        UNTIL NOT duplication;
        RETURN cid;
    END OpenNewOutputFile;

(***********************************************************************)
(*                        THE MAIN PROCEDURE                           *)
(***********************************************************************)

PROCEDURE TranslateOneFile (infile: FilenameString);

    (* Processes the GEDCOM file. *)

    VAR IS: InState;
        incid, outcid: ChanId;
        tempfile: FilenameString;

    BEGIN
        (* Open the input and output files. *)

        incid := OpenOldFile (infile, FALSE, FALSE);
        IF incid = NoSuchChannel THEN
            WriteString ("Sorry, file ");
            WriteString (infile);
            WriteString (" could not be opened.");
            WriteLn;
            RETURN;
        END (*IF*);
        SetPosition (incid, StartPosition(incid));
        IS := StartReading (incid);
        outcid := OpenNewOutputFile (tempfile);

        TranslateStream (IS, outcid);
        StopReading (IS);
        CloseFile (outcid);
        CloseFile (incid);

        (* The new version of the database is now in tempfile.  Delete  *)
        (* the original, and rename tempfile to the original file name. *)

        DeleteFile (infile);
        IF NOT MoveFile (tempfile, infile) THEN
            WriteString ("ERROR: Could not rename ");  WriteString (tempfile);
            WriteString (" to ");  WriteString (infile);  WriteLn;
        END (*IF*);

    END TranslateOneFile;

(************************************************************************)
(*                             MAIN PROGRAM                             *)
(************************************************************************)

PROCEDURE DoTheJob;

    (* Translates one or more GEDCOM files. *)

    VAR mask: FilenameString;
        D: DirectoryEntry;
        found: BOOLEAN;  pos: CARDINAL;

    BEGIN
        GetArgument (mask);
        IF mask[0] = Nul THEN
            WriteLn;
            WriteString ("Usage: ToUTF8 gedfilename");
            WriteLn;
            WriteLn;
            WriteString ("This will change the file's character set to UTF8.");
            WriteLn;
            WriteString ("Wildcards * and ? are allowed in the file name.");
            WriteLn;
            RETURN;
        END (*IF*);

        Strings.FindPrev ('.', mask, LENGTH(mask)-1, found, pos);
        IF NOT found THEN
            Strings.Append (".GED", mask);
        END (*IF*);


        found := FirstDirEntry (mask, FALSE, FALSE, FALSE, D);
        IF NOT found THEN
            WriteString ("Could not find any file matching ");
            WriteString (mask);  WriteLn;
        END (*IF*);

        WHILE found DO
            WriteString ("Translating file ");
            WriteString (D.name);  WriteLn;
            TranslateOneFile (D.name);
            found := NextDirEntry (D);
        END (*WHILE*);
        DirSearchDone (D);

    END DoTheJob;

(************************************************************************)

BEGIN
    NextName := "00000000";
    DoTheJob;
END ToUTF8.

