IMPLEMENTATION MODULE Transl8;

        (********************************************************)
        (*                                                      *)
        (*     Web Family Tree: part of the ToUTF8 utility      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            13 July 2024                    *)
        (*  Last edited:        30 December 2024                *)
        (*  Status:             Complete, untested              *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(* The purpose of this module is to translate text from several         *)
(* different character sets. The Gedcom standard, version 5.5, allows   *)
(* only ANSEL, UNICODE, or ASCII. It deprecates IBMPC as being          *)
(* ambiguous. UNICODE is also ambiguous, because it could be one of     *)
(* UTF8, two versions of UTF16 (bigendian and little endian), and two   *)
(* versions of UTF32. However, we handle all of those in subsidiary     *)
(* module InStream, which changes all Unicode variants to UTF8 on the   *)
(* run.  (There is also UTF-5 and UTF-7, but I don't think anyone uses  *)
(* them, so I don't support them.)                                      *)
(*                                                                      *)
(* Some GED software uses some of the "illegal" character sets anyway.  *)
(* so I choose to resolve the ambiguities as follows:                   *)
(*               ANSI       I assume it means Windows-1252              *)
(*               LATIN1     I assume it means ISO-8859-1                *)
(************************************************************************)

FROM SYSTEM IMPORT
    (* type *)  CARD16;

IMPORT Strings;

FROM InStream IMPORT
    (* type *)  Stream,
    (* proc *)  OpenStream, CloseStream, SReadLine, SReadRaw, StepBack;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  WriteRaw, FWriteString, FWriteLn;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST Nul = CHR(0);  CtrlZ = CHR(26);

TYPE
    CharSet = SET OF CHAR;
    cscode = (utf8, ansel, ansi, latin1, ascii);
    InState = POINTER TO
                  RECORD
                      cid: ChanId;
                      stream: Stream;
                      charset: cscode;
                  END (*RECORD*);

CONST Digits = CharSet {'0'..'9'};

(************************************************************************)
(*                  MISCELLANEOUS UTILITY PROCEDURES                    *)
(************************************************************************)

PROCEDURE HeadMatch (VAR (*INOUT*) string: ARRAY OF CHAR;
                              template: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff a leading substring of string matches template, *)
    (* modulo character case.  If there is a match, the matched         *)
    (* substring and any following spaces are removed from string.      *)

    VAR j: CARDINAL;

    BEGIN
        j := 0;
        LOOP
            IF (j > HIGH(template)) OR (template[j] = Nul) THEN
                EXIT (*LOOP*);
            END (*IF*);
            IF (j > HIGH(string)) OR (string[j] = Nul)
                    OR (CAP(string[j]) <> CAP(template[j])) THEN
                RETURN FALSE;
            END (*IF*);
            INC (j);
        END (*LOOP*);
        WHILE string[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        Strings.Delete (string, 0, j);
        RETURN TRUE;
    END HeadMatch;

(************************************************************************)
(*                 OPENING AND CLOSING THE INPUT STATE                  *)
(************************************************************************)

PROCEDURE StartReading (cid: ChanId): InState;

    (* Begins operations on the specified input channel. *)

    VAR IS: InState;

    BEGIN
        NEW (IS);
        IS^.cid := cid;
        IS^.stream := OpenStream(cid);
        IS^.charset := ascii;
        RETURN IS;
    END StartReading;

(************************************************************************)

PROCEDURE StopReading (VAR (*INOUT*) IS: InState);

    (* Closes down the operation. *)

    BEGIN
        CloseStream (IS^.stream);
        DISPOSE (IS);
    END StopReading;

(************************************************************************)
(*                  HANDLING THE INITIAL HEAD RECORD                    *)
(************************************************************************)

PROCEDURE ExtractLevel (VAR (*INOUT*) line: ARRAY OF CHAR): CARDINAL;

    (* Takes one line, trims its level from the front and returns the   *)
    (* level as the function result.                                    *)

    TYPE CharSet = SET OF CHAR;

    CONST Digits = CharSet {'0'..'9'};

    VAR j, level: CARDINAL;

    BEGIN
        level := 0;
        j := 0;
        WHILE line[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        IF line[j] IN Digits THEN
            REPEAT
                level := 10*level + (ORD(line[j]) - ORD('0'));
                INC (j);
            UNTIL NOT (line[j] IN Digits);
        ELSE
            level := MAX(CARDINAL);
        END (*IF*);
        WHILE line[j] = ' ' DO
            INC (j);
        END (*WHILE*);
        IF j > 0 THEN
            Strings.Delete (line, 0, j);
        END (*IF*);
        RETURN level;
    END ExtractLevel;

(************************************************************************)

PROCEDURE CopyHeadRecord (IS: InState;  outcid: ChanId);

    (* Copies the initial level 0 record from IS to outcid, and *)
    (* interprets (and modifies) the CHAR specification.        *)

    TYPE TableType = ARRAY cscode OF ARRAY [0..15] OF CHAR;

    CONST
        NameTable = TableType {"UNICODE", "ANSEL", "ANSI", "LATIN1", "ASCII"};

    VAR thisline, linecopy: ARRAY [0..255] OF CHAR;  level: CARDINAL;
        Charset: ARRAY [0..15] OF CHAR;
        pos: CARDINAL;  code: cscode;  found: BOOLEAN;

    BEGIN
        Charset[0] := Nul;
        SReadLine (IS^.stream, thisline);
        Strings.FindNext ("HEAD", thisline, 0, found, pos);
        IF found THEN
            LOOP
                FWriteString (outcid, thisline);  FWriteLn (outcid);
                SReadLine (IS^.stream, thisline);
                Strings.Assign (thisline, linecopy);
                level := ExtractLevel (linecopy);
                IF level = 0 THEN
                    StepBack (IS^.stream);
                    EXIT(*LOOP*);
                ELSIF (level = 1) AND HeadMatch (linecopy, "CHAR") THEN
                    Strings.Assign (linecopy, Charset);
                    FWriteString (outcid, "1 CHAR UNICODE");  FWriteLn (outcid);
                    EXIT (*LOOP*);
                END (*IF*);
            END (*LOOP*);
        ELSE
            StepBack (IS^.stream);
        END (*IF*);


        IF Charset[0] = Nul THEN code := ascii;
        ELSE
            code := MIN(cscode);  found := FALSE;
            REPEAT
                found := Strings.Equal (Charset, NameTable[code]);
                IF NOT found THEN
                    INC (code);
                END (*IF*);
            UNTIL found OR (code = MAX(cscode));
            IF NOT found THEN code := ascii END(*IF*);
        END (*IF*);

        IS^.charset := code;

    END CopyHeadRecord;

(************************************************************************)
(*                             UNICODE TO UTF8                          *)
(************************************************************************)

PROCEDURE UnicodeToUTF8 (code: CARDINAL;  VAR(*INOUT*) buffer: ARRAY OF CHAR;
                                            VAR (*INOUT*) k: CARDINAL);

    (* Stores the UTF8 form of code at buffer[k], updates k. *)

    VAR temp: CARDINAL;

    BEGIN
        IF code = 0 THEN

            (* Unknown code, replace it with '?' *)

            buffer[k] := '?';

        ELSIF code < 080H THEN
            (* A 7-bit number. *)

            buffer[k] := CHR(code);
            INC (k);

        ELSIF code < 0800H THEN
            (* An 11-bit number. *)

            temp := code DIV 16; (* A 7-bit number *)
            buffer[k] := CHR(0C0H + (temp DIV 4)); (* first 5 bits *)
            INC (k);
            buffer[k] := CHR(080H + 16*(temp MOD 4) + (code MOD 16));
                                                (* last 6 bits *)
        ELSE
            (* Up to 20 bits. *)

            temp := code DIV 16;  (* A 12-bit number *)
            buffer[k] := CHR(0E0H + (temp DIV 256)); (* first 4 bits *)
            INC (k);
            temp := temp MOD 256;  (* An 8-bit number *)
            buffer[k] := CHR(080H + (temp DIV 4)); (* next 6 bits *)
            INC (k);
            buffer[k] := CHR(080H + 16*(temp MOD 4) + (code MOD 16));
                                                (* last 6 bits *)
        END (*IF*);
        INC (k);

    END UnicodeToUTF8;

(************************************************************************)
(*                        DEALING WITH ANSEL                            *)
(************************************************************************)

CONST
    SpecialCaseTableSize = 53;

TYPE
    SpecialCaseTableType = ARRAY [0..SpecialCaseTableSize-1] OF
                               RECORD
                                   ch1, ch2: CHAR;
                                   unicode: CARD16;
                               END (*RECORD*);

CONST
    grave = CHR(0E1H);
    acute = CHR(0E2H);
    circum = CHR(0E3H);
    tilde = CHR(0E4H);
    umlaut = CHR(0E8H);
    ring = CHR(0EAH);
    cedilla = CHR(0F0H);

    SCT = SpecialCaseTableType{
                {grave, 'A', 00C0H},
                {grave, 'E', 00C8H},
                {grave, 'I', 00CCH},
                {grave, 'O', 00D2H},
                {grave, 'U', 00D9H},
                {grave, 'a', 00E0H},
                {grave, 'e', 00E8H},
                {grave, 'i', 00ECH},
                {grave, 'o', 00F2H},
                {grave, 'u', 00F9H},

                {acute, 'A', 00C1H},
                {acute, 'E', 00C9H},
                {acute, 'I', 00CDH},
                {acute, 'O', 00D3H},
                {acute, 'U', 00DAH},
                {acute, 'Y', 00DDH},
                {acute, 'a', 00E1H},
                {acute, 'e', 00E9H},
                {acute, 'i', 00EDH},
                {acute, 'o', 00F3H},
                {acute, 'u', 00FAH},
                {acute, 'y', 00FDH},

                {circum, 'A', 00C2H},
                {circum, 'E', 00CAH},
                {circum, 'I', 00CEH},
                {circum, 'O', 00D4H},
                {circum, 'U', 00DBH},
                {circum, 'a', 00E2H},
                {circum, 'e', 00EAH},
                {circum, 'i', 00EEH},
                {circum, 'o', 00F4H},
                {circum, 'u', 00FBH},

                {tilde, 'A', 00C3H},
                {tilde, 'N', 00D1H},
                {tilde, 'O', 00D5H},
                {tilde, 'a', 00E3H},
                {tilde, 'n', 00F1H},
                {tilde, 'o', 00F5H},

                {umlaut, 'A', 00C4H},
                {umlaut, 'E', 00CBH},
                {umlaut, 'I', 00CFH},
                {umlaut, 'O', 00D6H},
                {umlaut, 'U', 00DCH},
                {umlaut, 'a', 00E4H},
                {umlaut, 'e', 00EBH},
                {umlaut, 'i', 00EFH},
                {umlaut, 'o', 00F6H},
                {umlaut, 'u', 00FCH},
                {umlaut, 'y', 00FFH},

                {ring, 'A', 00C5H},
                {ring, 'a', 00E5H},

                {cedilla, 'C', 00C7H},
                {cedilla, 'c', 00E7H}
                };

(************************************************************************)

PROCEDURE SpecialCase (prefix, nextch: CHAR;
                               VAR (*OUT*) code: CARD16): BOOLEAN;

    (* Deals with accented characters that are represented in ANSEL by  *)
    (* a prefix followed by the unaccented character.  If we can        *)
    (* translate this to a single Unicode character, we return the      *)
    (* Unicode value in code and return TRUE.  A return value of FALSE  *)
    (* means that we're leaving the job to the caller.                  *)

    VAR k: CARDINAL;  found: BOOLEAN;

    BEGIN
        found := FALSE;
        IF ORD(prefix) >= 0E0H THEN
            k := 0;
            WHILE (k < SpecialCaseTableSize) AND NOT found DO
                IF SCT[k].ch1 > prefix THEN
                    k := SpecialCaseTableSize;
                ELSIF (SCT[k].ch1 < prefix) OR (SCT[k].ch2 < nextch) THEN
                    INC (k);
                ELSIF SCT[k].ch2 > nextch THEN
                    k := SpecialCaseTableSize;
                ELSE
                    code := SCT[k].unicode;
                    found := TRUE;
                END (*IF*);
            END (*WHILE*);
        END (*IF*);
        RETURN found;
    END SpecialCase;

(************************************************************************)

TYPE ANSELTransTable = ARRAY [CHR(0A0H)..CHR(0FFH)] OF CARD16;

CONST ANSELtoUnicode = ANSELTransTable{
          0000H, 0141H, 00D8H, 0110H, 00DEH, 00C6H, 0152H, 02B9H,
          00B7H, 266DH, 00AEH, 00B1H, 01A0H, 01AFH, 02BCH, 0000H,
          02BBH, 0142H, 00F8H, 0111H, 00FEH, 00E6H, 0153H, 02BAH,
          0131H, 00A3H, 00F0H, 0000H, 01A1H, 01B0H, 0000H, 0000H,
          00B0H, 2113H, 2117H, 00A9H, 266FH, 00BFH, 00A1H, 0000H,
          0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 00DFH,
          0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 0000H,
          0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 0000H, 0000H,
          0309H, 0300H, 0301H, 0302H, 0303H, 0304H, 0306H, 0307H,
          0308H, 030CH, 030AH, 0FE20H, 0FE21H, 0315H, 030BH, 0310H,
          0327H, 0328H, 0323H, 0324H, 0325H, 0333H, 0332H, 0326H,
          031CH, 032EH, 0FE22H, 0FE23H, 0000H, 0000H, 0313H, 0000H};

(************************************************************************)

PROCEDURE TranslateFromANSEL (VAR (*INOUT*) instr, outstr: ARRAY OF CHAR;
                                    amount: CARDINAL): CARDINAL;

    (* Translates up to amount bytes in instr from ANSEL to UTF-8.  *)
    (* Returns the number of bytes produced in outstr.              *)

    CONST LineLength = 1024;

    VAR j, k: CARDINAL;
        code, temp: CARD16;
        ch, prefixcode: CHAR;
        needsuffix: BOOLEAN;

    BEGIN
        j := 0;  k := 0;
        prefixcode := Nul;
        needsuffix := FALSE;
        WHILE (j < amount) AND (k < LineLength-3) AND (instr[j] <> Nul) DO
            temp := 0;
            IF needsuffix THEN
                ch := prefixcode;
                needsuffix := FALSE;
            ELSE
                ch := instr[j];  INC(j);
                IF ORD(ch) >= 0E0H THEN
                    IF NOT SpecialCase (ch, instr[j], temp) THEN
                        needsuffix := TRUE;
                        prefixcode := ch;
                        ch := instr[j];
                    END (*IF*);
                    INC (j);
                END (*IF*);
            END (*IF*);

            IF ORD(ch) = 0 THEN

                (* Special code for Nul. *)

                outstr[k] := CHR(0C0H);
                INC (k);
                outstr[k] := CHR(080H);
                INC (k);
            ELSIF ORD(ch) < 0A0H THEN
                outstr[k] := ch;
                INC (k);

            ELSE

                IF temp <> 0 THEN
                    code := temp;
                ELSE
                    code := ANSELtoUnicode[ch];
                END (*IF*);
                UnicodeToUTF8 (code, outstr, k)
            END (*IF*);

        END (*WHILE*);

        IF k < LineLength THEN
            outstr[k] := Nul;
        END (*IF*);

        RETURN k;

    END TranslateFromANSEL;

(************************************************************************)
(*                  TRANSLATION FROM ANSI AND LATIN1                    *)
(************************************************************************)

(* Remark: in the range 0A0H to 0FFH, Windows-1252 and Latin-1 and      *)
(* Unicode all agree with one another, so we only need a lookup table   *)
(* for the small range 080H to 09FH.                                    *)

TYPE UpperTable = ARRAY [CHR(080H)..CHR(09FH)] OF CARDINAL;

CONST undef = 0BFH;    (* arbitrary decision for undefined code points *)

CONST W1252toUnicode = UpperTable{
        020ACH,  undef, 0201AH, 00192H, 0201EH, 02026H, 02020H, 02021H,
        002C6H, 02030H, 00160H, 02039H, 00152H,  undef, 0017DH,  undef,
         undef, 02018H, 02019H, 0201CH, 0201DH, 02022H, 02013H, 02014H,
        002DCH, 02122H, 00161H, 0203AH, 00153H,  undef, 0017EH, 00178H};

(************************************************************************)

PROCEDURE TranslateFromWindows1252 (VAR (*IN*) inbuff: ARRAY OF CHAR;
                             VAR (*IN*) outbuff: ARRAY OF CHAR;
                               amount: CARDINAL): CARDINAL;

    (* Translates up to amount bytes from inbuff to outbuff, returns    *)
    (* the number of output bytes.                                      *)

    VAR code, j, k: CARDINAL;  ch: CHAR;

    BEGIN
        j := 0;  k := 0;
        WHILE j < amount DO
            ch := inbuff[j];  INC (j);
            IF (ch >= CHR(080H)) AND (ch <= CHR(0A0H)) THEN
                code := W1252toUnicode[ch];
            ELSE
                code := ORD(ch);
            END (*IF*);
            UnicodeToUTF8 (code, outbuff, k);
        END (*WHILE*);
        RETURN k;
    END TranslateFromWindows1252;

(************************************************************************)

PROCEDURE TranslateFromLatin1 (VAR (*IN*) inbuff: ARRAY OF CHAR;
                             VAR (*IN*) outbuff: ARRAY OF CHAR;
                               amount: CARDINAL): CARDINAL;

    (* Translates up to amount bytes from inbuff to outbuff, returns    *)
    (* the number of output bytes.                                      *)

    VAR code, j, k: CARDINAL;

    BEGIN
        j := 0;  k := 0;
        WHILE j < amount DO
            code := ORD(inbuff[j]);  INC (j);
            IF (code >= 080H) AND (code <= 0A0H) THEN
                DEC (code, 080H);
            END (*IF*);
            UnicodeToUTF8 (code, outbuff, k);
            INC (k);
        END (*WHILE*);
        RETURN k;
    END TranslateFromLatin1;

(************************************************************************)
(*                  THE OVERALL TRANSLATION OPERATION                   *)
(************************************************************************)

PROCEDURE Copy (VAR (*IN*) inbuff: ARRAY OF CHAR;  amount: CARDINAL;
                        VAR (*OUT*) outbuff: ARRAY OF CHAR): CARDINAL;

    (* Copies amount bytes, unaltered, from inbuff to outbuff.  *)

    VAR k: CARDINAL;

    BEGIN
        FOR k := 0 TO amount-1 DO
            outbuff[k] := inbuff[k];
        END (*FOR*);
        RETURN amount;
    END Copy;

(************************************************************************)

PROCEDURE Translate (cs: cscode;  VAR (*IN*) inbuff: ARRAY OF CHAR;
                      amount: CARDINAL;
                        VAR (*OUT*) outbuff: ARRAY OF CHAR): CARDINAL;

    (* Translates amount bytes from inbuff, puts the result in  *)
    (* outbuff, returns the number of bytes in outbuff.         *)

    BEGIN
        CASE cs OF
                utf8:   (* nothing to do *)
                        amount := Copy (inbuff, amount, outbuff);

            |   ansel:  amount := TranslateFromANSEL (inbuff, outbuff, amount);

            |   ansi:   amount := TranslateFromWindows1252 (inbuff, outbuff, amount);
                        WriteLn;

            |   latin1: amount := TranslateFromLatin1 (inbuff, outbuff, amount);
                        WriteLn;

            |   ascii:  (* nothing to do *)
                        amount := Copy (inbuff, amount, outbuff);

        END (*CASE*);
        RETURN amount;
    END Translate;

(************************************************************************)

PROCEDURE TranslateStream (IS: InState;  outcid: ChanId);

    (* Completes the translation from where StartReading left off.*)

    CONST buffersize = 128;

    VAR buffer, outbuff: ARRAY [0..buffersize-1] OF CHAR;
        amount: CARDINAL;
        done: BOOLEAN;

    BEGIN
        CopyHeadRecord (IS, outcid);
        REPEAT
            SReadRaw (IS^.stream, buffer, buffersize DIV 3, amount);
            done := (amount = 0) OR (buffer[0] = CtrlZ);
            IF NOT done THEN
                amount := Translate (IS^.charset, buffer, amount, outbuff);
                WriteRaw (outcid, outbuff, amount);
            END (*IF*);
        UNTIL done;
    END TranslateStream;

(************************************************************************)

END Transl8.

